#include <avr/interrupt.h>
#include <avr/io.h>

#include <ctype.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <util/delay.h>

#include "lcd.h"
#include "stack.h"

char keypad[16] = {'1', '2', '3', '/',
                   '4', '5', '6', '*',
                   '7', '8', '9', '-',
                   'C', '0', '=', '+'};

char a_string[16] = {0};
int a_digits = 0;
double a = 0, b = 0;
char op = ' ';
double result = 0;
stack s = {{0}, {0}, 0, 0};

uint8_t oper = 0;
uint8_t clear_display = 0;
/*
 * Rezultatul este cu virgula.
 */
uint8_t floating_point = 0;

uint8_t total_overflow = 0;
/*
 * Intreruperea de overflow pentru TIMER1.
 */
ISR(TIMER1_OVF_vect)
{
    ++total_overflow;

    if (total_overflow >= 2) {
        if (!bit_is_set(PINB, PB4))
            PORTB = 0b11011111;
        else if (!bit_is_set(PINB, PB5))
            PORTB = 0b10111111;
        else if (!bit_is_set(PINB, PB6))
            PORTB = 0b01111111;
        else if (!bit_is_set(PINB, PB7))
            PORTB = 0b11101111;

        total_overflow = 0;
    }
}

/*
 * Initializeaza TIMER1.
 */
void timer1_init()
{
    /*
     * Modul Normal (TOP = 0xffff), prescaler 8.
     */
    TCCR1B = (1 << CS11);

    /*
     * Activeaza intreruperea de overflow.
     */
    TIMSK1 |= (1 << TOIE1);
}

/*
 * Initializeaza keypad-ul si led-ul conectat la PD7 ce se va aprinde cand
 * este apasata o tasta.
 */
void IO_init()
{
    // butoanele sunt puse ca intrare, cu rezistenta de pull-up.
    /*
     * Pinii PB0-PB3 sunt folositi pt a selecta liniile si sunt setati ca input.
     * Pinii PB4-PB7 sunt folositi pt a selecta coloanele si sunt setati ca output.
     * Initial PB4(coloana 1) este la 0V pentru a lua input doar de pe ea.
     * Coloanele sunt puse la 0V pe rand in intreruperea de overflow pentru TIMER1.
     */
    DDRB = 0b11110000;
	PORTB = 0b11101111;

    /*
     * PD7 (led-ul) ca pin de ieșire.
    */
    DDRD |= (1 << PD7);
}

void calculator_init() {
    s = {{0}, {0}, 0, 0};
    memset(a_string, 0, 10);
    a_digits = 0;
    op = ' ';
    result = 0;
    floating_point = 0;
    clear_display = 0;
    oper = 0;
}

void print_key_to_LCD(char key) {
    if (key == 'C') {
        LCD_writeInstr(LCD_INSTR_clearDisplay);
        calculator_init();
        return ;
    } else if (key == '=') {
        LCD_writeInstr(LCD_INSTR_nextLine);
        LCD_putChar(key);
        char str[16] = {0};
        if(floating_point == 1)
            sprintf(str, "%.2f", result);
        else
            sprintf(str, "%.0f", result);
        LCD_print(str);
        return ;
    }

    LCD_putChar(key);
}

/*
 * Verifica daca a fost apasata o tasta de pe keypad si ii returneaza indexu-ul
 * din matricea tinuta ca vector keypad[16].
 * Deoarece functia este apelata intr-un while(1), returnez -1 cand nu
 * este apasata nici o tasta.
 */
int8_t get_keypad_index() {
    for (uint8_t i = 0; i < 4; ++i) {
        if (!bit_is_set(PINB, i)) {
            for (uint8_t j = 0; j < 4; ++j) {
                if (!bit_is_set(PINB, j + 4))
                    return i * 4 + j;
            }
        }
    }

    return -1;
}

/*
 * Functie ce verifica precedenta operatorilor.
 * Intai inmultirile si impartirile, dupa care minus si dupa aceea plus.
 */
uint8_t compare_precedence(char op1, char op2)
{
    if (op1 == '*' || op1 == '/' || (op1 == '-' && (op2 == '+' || op2 == '-')))
        return 1;
    return 0;
}

/*
 * Functie ce verifica daca tastele sunt blocate si se asteapta o anumita
 * tasta pentru a le debloca si pe restul.
 */
void check_blocked() {
    if (clear_display == 1) {
        while (1) {
            int8_t key = get_keypad_index();
            if (keypad[key] == 'C')
                return;
        }
    }

    else if (oper == 1) {
        while (1) {
            int8_t key = get_keypad_index();
            if (key != -1 && (isdigit(keypad[key]) || keypad[key] == 'C')) {
                oper = 0;
                return;
            }
        }
    }
}

/*
 * Functie ce ia 2 operanzi si un operator din stivele fiecaruia
 * si pune result = a op b inapoi in stiva operanzilor.
 */
void store_temp_result() {
    op = stack_pop_operator(&s);
    b = stack_pop_operand(&s);
    a = stack_pop_operand(&s);

    result = 0;
    switch (op) {
        case '/' :
            if(fmod(a, b) != 0)
                floating_point = 1;
            result = a / b;
            break;
        case '*' :
            result = a * b;
            break;
        case '-' :
            result = a - b;
            break;
        default :
            result = a + b;
    }
    stack_push_operand(&s, result);
}

/*
 * Adauga inca o cifra la sfarsitul operand-ului reprezentate deocamdata ca un char[].
 */
void get_digit(char key) {
    a_string[a_digits++] = key;
}

/*
 * Functia este apelata in functia get_operator() si get_result(),
 * deoarece atunci stim ca operand-ul este "complet".
 * Functie transforma operand-ul din char[] intr-un double
 * pe care il salveaza in stiva operanzilor.
 */
void store_operand() {
    a = strtol(a_string, NULL, 10);
    memset(a_string, 0, 10);
    a_digits = 0;

    stack_push_operand(&s, a);
    a = 0;
}

/*
 * Functia este apelata atunci cand a fost introdus un operator la tastatura.
 * Algoritmul functiei este in felul urmator:
 * Verific daca este primul operator introdus sau daca are o prioritate mai mare
 * decat varful-ul stivei. In acest caz doar introduc operatorul in stiva.
 * Daca operator-ul introdus are o prioritate mai mica, atunci inainte de a il
 * introduce in stiva, scot varf-ul stivei si fac operatie pe cei 2 operanzi
 * scosi din stiva operanzilor, iar rezultatul il introduc in stiva operanzilor,
 * cu ajutorul functiei store_temp_result().
 * Acum verific din nou daca operatorul care fusese introdus de la tastatura
 * are prioritatea mai mica decat varf-ul stivei. Daca da se repeta pasii anteriori.
 * Daca nu, atunci este insfarsit introdus in stiva operatorilor.
 */
void get_operator(char key) {
    /*
     * Verifica ca prima data sa nu fie un operator fara nici un operand pe LCD.
     */
    if (a_digits == 0 || (a_digits == 1 && a_string[a_digits - 1] == '-')) {
        PORTD &= ~(1 << PD7);
        return ;
    }

    /*
     * Variabila folosita pentru a bloca tastele pt operatori, sa nu am
     * mai mult de un operator pe LCD si sa nu fie mai multi operatori in stiva
     * lor pentru a nu strica algoritmul.
     */
    oper = 1;

    store_operand();

    if (stack_size_operator(&s) == 0) {
        stack_push_operator(&s, key);
    } else if (compare_precedence(stack_peek_operator(&s), key)) {
        while (compare_precedence(stack_peek_operator(&s), key))
            store_temp_result();
        stack_push_operator(&s, key);
    } else if (!compare_precedence(stack_peek_operator(&s), key)) {
        stack_push_operator(&s, key);
    }
}

/*
 * Functia este apelata cand este introdusa tasta '='.
 * scot din varf-ul stivei operatorilor si fac operatie pe cei 2 operanzi
 * scosi din stiva operanzilor, iar rezultatul il introduc in stiva operanzilor,
 * cu ajutorul functiei store_temp_result(). Fac asta cat timp stiva operatorilor
 * nu este goala.
 * Cand stiva este goala, rezultatul este salvat in variabila globala result.
 */
void get_result() {
    store_operand();

    while (stack_size_operator(&s) > 0)
        store_temp_result();

    /*
     * Variabila folosita pentru a bloca toate tastele, inafara de tasta 'C'
     * deoarece a terminat de evaluat expresia matematica, iar LCD-ul trebuie curatat.
     */
    clear_display = 1;
}

int main()
{
    IO_init();
    LCD_init();
    sei();
    timer1_init();
    calculator_init();

    LCD_print("Calculator");
    LCD_writeInstr(LCD_INSTR_nextLine);
    LCD_print("Radu Neagoe");
    clear_display = 1;

    while (1) {
        /*
         * Verific daca am deja rezultatul si trebuie sa curat display-ul,
         * sau daca am deja un operator pe LCD si trebuie sa introduc un numar.
         */
        check_blocked();

        /*
         * A fost apasata o tasta de pe keypad.
         */
        int8_t key = get_keypad_index();
        if (key != -1) {
            /*
             * Aprinde led.
             */
            PORTD |= (1 << PD7);

            /*
             * Verific ce tasta a fost apasata.
             *
             * A doua conditie din primul if este pentru cazul in care primul numar vreau sa fie negativ.
             */
            if (isdigit(keypad[key]) || (a_digits == 0 && keypad[key] == '-')) {
                get_digit(keypad[key]);
            } else if (keypad[key] == '/' || keypad[key] == '*' || keypad[key] == '+' || keypad[key] == '-') {
                get_operator(keypad[key]);
                if (oper == 0)
                    continue;
            } else if (keypad[key] == '=') {
                get_result();
            }

            /*
             * Afisez caracterul primit de la keypad pe LCD.
             */
            print_key_to_LCD(keypad[key]);

            /*
             * Delay-ul este necesar pentru a evita debouncing-ul de la taste.
             */
            _delay_ms(100);

            /*
             * Stinge led.
             */
            PORTD &= ~(1 << PD7);
        }
    }

    return 0;
}
